#include <stdio.h>     // ISO C Standard: 4.9 INPUT/OUTPUT
#include <time.h>      // ISO C Standard: 4.12 DATE and TIME 
#include <unistd.h>    // POSIX Standard: 2.10 Symbolic Constants  
#include <fcntl.h>     // POSIX Standard: 6.5 File Control Operations
#include <sys/types.h> // POSIX Standard: 2.6 Primitive System Data Types
#include <sys/stat.h>  // POSIX Standard: 5.6 File Characteristics
#include <string.h>  
#include <iostream>  
#include <sys/time.h>
#include <stdarg.h> //for Putf
#include "TLSerial.h"

TLSerial::TLSerial(const Int_t bufsize):TLconnect(bufsize)
{
  fSerOK=true;
}

TLSerial::~TLSerial()
{
  Close();
  delete [] fDevice;
  delete fTerm;
}

/*
Int_t TLSerial::Init(const Char_t* device,const Char_t* conffile,bool(*setup)(void*)){
  fTerm=new termios;
  fDevice=new Char_t[strlen(device)+1];
  strcpy(fDevice,device);
  if ((fPort  = open( fDevice, O_RDWR|O_NOCTTY|O_NONBLOCK ))== -1){ 
    Message(0,"Error:Can't open comm port %s",fDevice);
    fSerOK = false;
    return -1;
  }
  termios term;
  bzero(&term, sizeof(term));

  if (tcgetattr(fPort, fTerm) == -1) {
    Message(0,"Can't get device settings(tcgetattr)");
    fSerOK = false;
    return -2;
  }
  if (setup!=NULL && !setup((void*)&term)){ //call external setup method
    Message(3,"Using internal setup");
    //  term.c_iflag = IXOFF;  
    term.c_iflag = 0;  
    //  term.c_oflag = OLCUC | IXON  ; 
    term.c_oflag = 0; 
    //  term.c_cflag = B9600|CS8|CREAD|CLOCAL;
    term.c_cflag = B9600|CS8|CRTSCTS|CREAD;
    //  term.c_cflag = B19200|CS8|CREAD;
    //term.c_cflag = B115200|CS8|CREAD;
    term.c_lflag = 0;
    term.c_cc[VINTR]    = 3;   // Ctrl-c END OF TEXT 
    term.c_cc[VQUIT]    = 28;  // Ctrl-\ Quit character 
    term.c_cc[VERASE]   = 127; // del 
    term.c_cc[VKILL]    = 21;  // Ctrl-u  
    term.c_cc[VEOF]     = 4;   // Ctrl-d 
    term.c_cc[VTIME]    = 0;   // inter-character timer unused 
    term.c_cc[VMIN]     = 1;   // blocking read until 1 character arrives 
    term.c_cc[VSWTC]    = 0;   // '\0' 
    term.c_cc[VSTART]   = 17;  // Ctrl-q  
    term.c_cc[VSTOP]    = 19;  // Ctrl-s 
    term.c_cc[VSUSP]    = 26;  // Ctrl-z 
    term.c_cc[VEOL]     = 0;   // '\0' 
    term.c_cc[VREPRINT] = 18;  // Ctrl-r 
    term.c_cc[VDISCARD] = 15;  // Ctrl-o 
    term.c_cc[VWERASE]  = 23;  // Ctrl-w 
    term.c_cc[VLNEXT]   = 22;  // Ctrl-v 
    term.c_cc[VEOL2]    = 0;   // '\0' 
  } else {
    Message(3,"Using external setup");
  }
  
  tcflush(fPort,TCIFLUSH);

  if (tcsetattr( fPort, TCSANOW, &term )) {
    Message(0,"Bad parameters for port %s",fDevice);
    fSerOK = false;
    return -3;
  }

  return TLconnect::Init(device,conffile);
}
*/
Int_t TLSerial::Init(const Char_t* device,const Char_t* conffile){
  fTerm=new termios;
  fDevice=new Char_t[strlen(device)+1];
  strcpy(fDevice,device);
  if ((fPort  = open( fDevice, O_RDWR|O_NOCTTY|O_NONBLOCK ))== -1){ 
    Message(0,"Error:Can't open comm port %s",fDevice);
    fSerOK = false;
    return -1;
  }
  termios term;
  bzero(&term, sizeof(term));

  if (tcgetattr(fPort, fTerm) == -1) {
    Message(0,"Can't get device settings(tcgetattr)");
    fSerOK = false;
    return -2;
  }
  if (1){ //call external setup method
    Message(3,"Using internal setup");
    //  term.c_iflag = IXOFF;  
    term.c_iflag = 0;  
    //  term.c_oflag = OLCUC | IXON  ; 
    term.c_oflag = 0; 
    //  term.c_cflag = B9600|CS8|CREAD|CLOCAL;
    term.c_cflag = B9600|CS8|CRTSCTS|CREAD;
    //  term.c_cflag = B19200|CS8|CREAD;
    //term.c_cflag = B115200|CS8|CREAD;
    term.c_lflag = 0;
    term.c_cc[VINTR]    = 3;   // Ctrl-c END OF TEXT 
    term.c_cc[VQUIT]    = 28;  // Ctrl-\ Quit character 
    term.c_cc[VERASE]   = 127; // del 
    term.c_cc[VKILL]    = 21;  // Ctrl-u  
    term.c_cc[VEOF]     = 4;   // Ctrl-d 
    term.c_cc[VTIME]    = 0;   // inter-character timer unused 
    term.c_cc[VMIN]     = 1;   // blocking read until 1 character arrives 
#ifdef VSWTC
    term.c_cc[VSWTC]    = 0;   // '\0' 
#endif
    term.c_cc[VSTART]   = 17;  // Ctrl-q  
    term.c_cc[VSTOP]    = 19;  // Ctrl-s 
    term.c_cc[VSUSP]    = 26;  // Ctrl-z 
    term.c_cc[VEOL]     = 0;   // '\0' 
    term.c_cc[VREPRINT] = 18;  // Ctrl-r 
    term.c_cc[VDISCARD] = 15;  // Ctrl-o 
    term.c_cc[VWERASE]  = 23;  // Ctrl-w 
    term.c_cc[VLNEXT]   = 22;  // Ctrl-v 
    term.c_cc[VEOL2]    = 0;   // '\0' 
  } else {
    Message(3,"Using external setup");
  }
  
  tcflush(fPort,TCIFLUSH);

  if (tcsetattr( fPort, TCSANOW, &term )) {
    Message(0,"Bad parameters for port %s",fDevice);
    fSerOK = false;
    return -3;
  }

  return TLconnect::Init(device,conffile);
}


Int_t TLSerial::Send(const Char_t* data)
{
  Int_t nBytes;
  nBytes=PutRaw(data,strlen(data));
  if (nBytes<1) return nBytes;
  if (strlen(fstop)>0)
    nBytes+=PutRaw(fstop,strlen(fstop));
  Message(3,"Sent:%s %d bytes",data,nBytes);
  return nBytes;
}

Int_t TLSerial::Send(const Char_t* data,const Int_t length){
  Message(3,"Sending:first %d bytes of >%s<",length,data);
  return PutRaw(data,length);
}

Int_t TLSerial::PutRaw(const Char_t* data,const Int_t len)
{
  size_t nBytes=0;
  if (len==0) return 0;
  //  iSerOK =  nBytes=write(fPort,(const void *)data,len);
  //  iSerOK += nBytes=write(fPort,"\r",1);
  for (Int_t i=0;i<len;i++){
    nBytes+=write(fPort,&data[i],1);
    usleep(10);
  }
  if (!nBytes) {
    Message(0,"Error:Problem writing to serial port");
    fSerOK = false;
    return -nBytes;
  }
  /*
  if (fdebug){
    printf("Send:");
    if (len==1 && data[0]==13) printf("CR");
    else if (len==1 && data[0]==10) printf("LF");
    else
      for (Int_t i=0;i<iSerOK;i++) printf("%c",data[i]);
    printf("\n");
  }
  */
  return nBytes;
}



Int_t TLSerial::Receive(void* output, const Int_t bufsize,Int_t timeout){
  // Inputs a line from the serial port, stopping at
  // end of line, carriage return, timeout, or buffer full.
  // The telescope sends always two LF (Char_t=10) at the end.
  // This was because of the input modes (input processing)
  //timeout in miliseconds
  Char_t* buf=(Char_t*) output;
  Int_t bufpos = 0;
  buf[bufpos] = 0;    // initialize empty string
  Int_t bufposmax = bufsize-1;  // maximum subscript
  Char_t c = 0;
  Int_t success = 0;
  Message(4,"Waiting to receive");
  GetChar(&c,&success,timeout);
  if (fdebug>3) std::cerr<<"Get:";
  if(success<=0) {
    if (fdebug>3) std::cerr<<std::endl;
    Message(1,"Error:SerGet Failed!");
    for(Int_t i=0;i<bufposmax;i++)
      buf[i]='X';
    buf[bufsize-1]=0;
    return bufpos;
  }
  
  Int_t stop_bits_received=0;
  while (1) {
    if (success <= 0) {     // timeout
      if (fdebug>3) std::cerr<<std::endl;
      Message(1,"Timed out waiting for serial input");
      fSerOK=false;
      return bufpos;
    }
    if (c == fstop[stop_bits_received]) {
      stop_bits_received++;
      if (strlen(fstop)==(size_t)stop_bits_received){
	if (fdebug>3) std::cerr<<"[STOP STRING]"<<std::endl;
	buf[bufpos]=0;//terminate string
	Message(3,"Received:%s",buf);
	return bufpos;
      } 
    } else if (stop_bits_received>0){//previous was not stopbit!
      for (Int_t i=0;i<stop_bits_received;i++){//refill buffer
	buf[bufpos] = fstop[i];
	bufpos++;    
	buf[bufpos] = 0;  // guarantee validly terminated string
	if (bufpos >= bufposmax) {
	  Message(0,"Error:Buffer full while refilling buffer");
	  return bufpos; 
	}
      }
      stop_bits_received=0;//reset counter
    }
    if (bufpos >= bufposmax) {
      Message(0,"Error:Buffer full, increase buffer size");
      return bufpos;
    }
    if (!stop_bits_received && c!=0){
      //if (!stop_bits_received){
      buf[bufpos] = c;
      bufpos++;
      buf[bufpos] = 0;  // guarantee validly terminated string
      if (fdebug>3) std::cerr<<c;//std::endl;
    }
    if (fdebug>3 && c==0) std::cerr<<"[ZERO STRING]";//std::endl;
    GetChar(&c,&success,timeout);
    //    if (fdebug && success) std::cerr<<c;//std::endl;
  }
}

void TLSerial::GetChar(Char_t* c, Int_t* success, Int_t timeout)
{
  //usleep(5000);//pause
  time_t finish;
  finish = time(NULL) + 5;
  *success = 0;
  size_t nBytes = 0;
  
  fd_set rfds;
  struct timeval tv;
  Int_t retval;
  FD_ZERO(&rfds);
  FD_SET(fPort, &rfds);
  // timeout in msec
  tv.tv_sec = timeout/1000;
  tv.tv_usec = (timeout%1000)*1000;

  retval = select(fPort+1, &rfds, NULL, NULL, &tv);
  if(retval) {
    nBytes=read(fPort,c ,(size_t)1);
    *success = nBytes;
  }
}

Int_t TLSerial::ReceiveRaw(void* output, const Int_t bufsize,Int_t timeout){
  // used for binary data transfer
  // Stops at timeout or requested bufsize
  // return the number of received bytes
  Char_t* buf=(Char_t*) output;
  Char_t c = 0;
  Int_t success = 0;
  
  for(Int_t i=0;i<bufsize;i++){
    GetChar(&c,&success,timeout);
    if(success<=0) {
      if (!i) {
	Message(1,"Error:Timed out waiting for serial input");  
	fSerOK=false;
	return 0; 
      } 
      return i+1;
    }
    buf[i] = c;
  }
  return bufsize;
}



Int_t TLSerial::Close()
{
  if (tcsetattr( fPort, TCSANOW, fTerm )){
    Message(0,"Error:Bad parameters for closing port %s",fTerm);
    fSerOK=false;
    return 0;
  } 
  fSerOK = !(close(fPort));
  if (!fSerOK) {
    Message(0,"Error:Problem closing serial port");
    return 0;
  }
  Message(3,"Serial port %s successfuly closed",fDevice);
  return 1;
}

void TLSerial::Message(Int_t level, const Char_t* msg2, ...)
{
  if (level>fdebug) return;
  std::cerr<<"TLSerial:(";

  Char_t msg3[1024];
  va_list ap;
  va_start(ap,msg2);
  vsprintf(msg3,msg2,ap);
  va_end(ap);

  std::cerr<<msg3<<")"<<std::endl;
}



Int_t TLSerial::CheckError(const Char_t* message){
  if (fSerOK){
    if (message) Message(1,message);
    return 1;
  } 
  if(message) Message(0,message);
  return 0;
}

/*
TLconnect* TLSerial::OpenPort(const Char_t* strport,Int_t debug,Bool_t (*setup)(void*)){
  //open port
  //strport must be specified in "serial:filename" e.g. "serial:/dev/ttyS0"
  if (!strcmp(strport,"help")) { 
    printf("serial\tRS232 protocol\n");
    return 0;
  }
  TLconnect* port;
  port=0;

  Int_t n=255;
  Char_t *p;
  Char_t *tmp=new Char_t[n+1];
  strncpy(tmp,strport,n+1);
  p = index(tmp,':');
  if (p) p[0]=0;//terminate string
  else return 0;

  if (strcmp(tmp,"serial")!=0) return 0;
  strncpy(tmp,strport,n+1);
  p = index(tmp,':')+1;
  if (p) {
    port=(TLconnect*)new TLSerial(1024);    
    port->fdebug=debug;
    if(port->Init(p,"",setup)<1) {
      port->Message(0,"Sorry, but %s port could not be opened.",strport); 
      return 0;
    }
  } 
  return port;
}                          

*/
TLconnect* TLSerial::OpenPort(const Char_t* strport,Int_t debug){
  //open port
  //strport must be specified in "serial:filename" e.g. "serial:/dev/ttyS0"
  if (!strcmp(strport,"help")) { 
    printf("serial\tRS232 protocol\n");
    return 0;
  }
  TLconnect* port;
  port=0;

  Int_t n=255;
  Char_t *p;
  Char_t *tmp=new Char_t[n+1];
  strncpy(tmp,strport,n+1);
  p = index(tmp,':');
  if (p) p[0]=0;//terminate string
  else return 0;

  if (strcmp(tmp,"serial")!=0) return 0;
  strncpy(tmp,strport,n+1);
  p = index(tmp,':')+1;
  if (p) {
    port=(TLconnect*)new TLSerial(1024);    
    port->fdebug=debug;
    if(port->Init(p,"")<1) {
      port->Message(0,"Sorry, but %s port could not be opened.",strport); 
      return 0;
    }
  } 
  return port;
}                          


/****************MODBUS************************/
/*
TLModbus::TLModbus(const Int_t bufsize):TLSerial(bufsize){
  
}

TLModbus::~TLModbus(){

}

void TLModbus::Message(const Char_t* msg1, const Char_t* msg2, ...)
{
  if (!strcmp(msg1,"Info:") && !fdebug) return;
  std::cerr<<"TLModbus:"<<msg1<<"(";

  Char_t msg3[1024];
  va_list ap;
  va_start(ap,msg2);
  vsprintf(msg3,msg2,ap);
  va_end(ap);

  std::cerr<<msg3<<")"<<std::endl;
}


TLconnect* TLModbus::OpenPort(const Char_t* strport,Int_t debug,Bool_t (*setup)(void*)){
  //open port
  //strport must be specified in "modbus:filename" e.g. "modbus:/dev/ttyS0"
  if (!strcmp(strport,"help")) { 
    printf("modbus\tRS232 modbus protocol\n");
    return 0;
  }
  TLconnect* port;
  port=0;

  Int_t n=255;
  Char_t *p;
  Char_t *tmp=new Char_t[n+1];
  strncpy(tmp,strport,n+1);
  p = index(tmp,':');
  if (p) p[0]=0;//terminate string
  else return 0;

  if (strcmp(tmp,"modbus")!=0) return 0;
  strncpy(tmp,strport,n+1);
  p = index(tmp,':')+1;
  if (p) {
    port=(TLconnect*)new TLModbus(1024);    
    port->fdebug=debug;
    if(port->Init(p,"",setup)<1) {
      port->Message("Error:","Sorry, but %s port could not be opened.",strport); 
      return 0;
    }
  } 
  return port;
}                          

*/
