#include "TLconnect.h"
#include <unistd.h>
#include <string.h>
#include <stdio.h>

TLconnect::TLconnect(const Int_t bufsize) {
  if (bufsize<1) fbuflen=32;
  else fbuflen=bufsize;
  fbuffer=new Char_t[fbuflen];
  fdebug=0;
}

TLconnect::~TLconnect() {
  if (fbuffer) delete [] fbuffer;
  else Message(0,"fbuffer is zero, that is strange!");
  fbuffer=0;
  fbuflen=0;
}

void
TLconnect::SetBufferSize(const Int_t bufsize){
  //resize the buffer 
  if (bufsize==fbuflen) return;
  if (fbuffer) delete [] fbuffer;
  fbuflen=bufsize;
  if (fbuflen>0)
    fbuffer=new Char_t[fbuflen];
  else Message(2,"Want to set buffer size to %d, ignoring!",bufsize);
}

int
TLconnect::Init(const Int_t address,const Char_t* conffile){
  if(strcmp(conffile,""))
    return Configure(conffile);
  return 1;
}

int
TLconnect::Init(const Char_t* address,const Char_t* conffile){
  if(strcmp(conffile,""))
    return Configure(conffile);
  return 1;
}

int
TLconnect::Close(){
  return 1;
}


Int_t TLconnect::Send(const Char_t* command){
  std::cout<<"Send:"<<command<<std::endl;
  return 1;
}

Int_t TLconnect::Send(const Char_t* command,const Int_t length){
  std::cout<<"Send:"<<command<<std::endl;
  return length;
}

Int_t TLconnect::Sendf(const Char_t* fmt, ...){
  va_list ap;
  va_start(ap,fmt);
  Char_t stmp[256];
  //cout<<"Ap0: "<<fmt<<endl;
  //Char_t* s=va_arg(ap,Char_t*);
  //Int_t s=va_arg(ap,Int_t);
  //cout<<"AP "<<s<<endl;
  vsprintf(stmp,fmt,ap);
  va_end(ap);
  return Send(stmp);
}

Int_t TLconnect::Receive(void* output, const Int_t length,Int_t timeout){
  return 0;//nothing received
}

Int_t TLconnect::ReceiveRaw(void* output, const Int_t length,Int_t timeout){
  return 0;//nothing received
}

Int_t TLconnect::Receive(const Int_t length,Int_t timeout){
  //receive into buffer
  if (!length) 
    return Receive(fbuffer,fbuflen,timeout);
  else 
    return Receive(fbuffer,length,timeout);
}

Int_t TLconnect::ReceiveRaw(const Int_t length,Int_t timeout){
  //receive into buffer
  return ReceiveRaw(fbuffer,length,timeout);
}

Int_t TLconnect::Query(const Char_t* send, Char_t* recv, Int_t size,Int_t timeout){
  if (Send(send))
    return Receive(recv,size,timeout);
  return 0;
}

Int_t TLconnect::Query(const Char_t* send,Int_t timeout){
  if (Send(send))
    return Receive(fbuflen,timeout);
  return 0;
}

Bool_t TLconnect::sQuery(const Char_t* send, Double_t& value){
  //used to query for single Double_t value
  //return the result into value and true if successful
  if (Query(send)<1) return false;
  if (sscanf(fbuffer,"%le",&value)==1) return true;
  Message(0,"Query(%s) unrecognized answer. See debug...",send);
  return false;
}

Bool_t TLconnect::sQuery(const Char_t* send, Int_t& value){
  //used to query for single Int_t value
  //return the result into value and true if successful
  if (Query(send)<1) return false;
  if (sscanf(fbuffer,"%d",&value)==1) return true;
  Message(0,"Query(%s) unrecognized answer. See debug...",send);
  return false;
}

Bool_t TLconnect::sQuery(const Char_t* send, UInt_t& value){
  //used to query for single Int_t value
  //return the result into value and true if successful
  if (Query(send)<1) return false;
  if (sscanf(fbuffer,"%d",&value)==1) return true;
  Message(0,"Query(%s) unrecognized answer. See debug...",send);
  return false;
}

Bool_t TLconnect::sQuery(const Char_t* send, Bool_t& value){
  //used to query for single Bool_t value
  //try to locate strings ON/OFF, TRUE/FALSE or 1/0
  //return the result into value and true if successful
  if (Query(send)<1) return false;
  if (strcasestr(fbuffer,"ON")){
    value=true;
    return true;
  }
  if (strcasestr(fbuffer,"OFF")){
    value=false;
    return true;
  }
  if (strstr(fbuffer,"1")){
    value=true;
    return true;
  }
  if (strcasestr(fbuffer,"0")){
    value=false;
    return true;
  }
  if (strstr(fbuffer,"TRUE")){
    value=true;
    return true;
  }
  if (strcasestr(fbuffer,"FALSE")){
    value=false;
    return true;
  }
  Message(0,"Query(%s) unrecognized answer. See debug...",send);
  return false;
}

Bool_t TLconnect::WaitOPC(Int_t timeout){
  //used to query Operation completed
  //it will wait forever 
  fbuffer[0]=0;
  Send("*OPC?"); //sometimes keithley respond, but sometimes it doesnt!
  while (fbuffer[0]!='1') {
    if (Receive(timeout)>0){
      if (fbuffer[0]=='1') break;
      Send("*OPC?");//resend the query
    }
    Message(1,"QueryOPC(%d) recognized timeout - ignoring and waiting.",timeout);
  }
  return true;
}


const Char_t* TLconnect::GetBuffer() const{
  return fbuffer;
}

void TLconnect::ClearBuffer(){
  if (fbuflen>0){
    for(Int_t i=0;i<fbuflen;i++)
      fbuffer[i]=0;
  }
}

void TLconnect::Message(const Char_t* msg1, const Char_t* msg2,...){
  std::cerr<<"TLconnect:"<<msg1<<"(";

  Char_t msg3[1024];
  va_list ap;
  va_start(ap,msg2);
  vsprintf(msg3,msg2,ap);
  va_end(ap);

  std::cerr<<msg3<<")"<<std::endl;
}

void TLconnect::Message(Int_t msg1, const Char_t* msg2,...){

  if (fdebug<msg1) return; 
  switch(msg1){
  case 0:std::cerr<<"Error:";break;
  case 1:std::cerr<<"Info:";break;
  case 2:std::cerr<<"Debug:";break;
  case 3:std::cerr<<"Debug+:";break;
  default:std::cerr<<"Debug++:";break;
  }
  std::cerr<<"(";

  Char_t msg3[2048];
  va_list ap;
  va_start(ap,msg2);
  vsprintf(msg3,msg2,ap);
  va_end(ap);

  std::cerr<<msg3<<")"<<std::endl;
  if (fdebug>255) sleep(fdebug-255);//wait 3s before proceeding
}


Int_t TLconnect::Configure(const Char_t* conffile){
  FILE* fp=fopen(conffile,"r");
  if(!fp) {
    Message("Error opening conf.file",conffile);
    return 0;
  }
  Char_t str[256];
  Int_t i=1;
  while(fgets(str,256,fp)) {
    if(strncmp(str,"//",2) && strncmp(str,"#",1))
      i+=Send(str);
  }
  fclose(fp);
  if (i>1) i--; 
  return i; 
}

Int_t TLconnect::CheckError(const Char_t* message){
  return 1;
}

const Char_t* TLconnect::GetPortStr(const Char_t* devport){
  //returns port string from the template device:port
  static Char_t strport[256];
  if (!devport || !strlen(devport)) return 0;
  if (sscanf(devport,"%*[^:]:%s",strport)==1) return strport;
  else return 0;
}
